use std::collections::HashMap;
use std::io;
use std::io::Read;
use std::net::{SocketAddr, TcpListener};
use std::os::fd::{AsRawFd, RawFd};
use messages::client_registration::ClientRegistration;
use messages::Serializable;
use crate::client::Client;

use crate::epoll::{Epoll, is_read_event, is_write_event};

// Based on: https://www.zupzup.org/epoll-with-rust/index.html

const KEY_NEW_CONNECTION: u16 = u16::MAX;

pub struct TcpServer {
    addr: SocketAddr,
    listener: TcpListener,
    listener_fd: RawFd,
    epoll: Epoll,
    request_contexts: HashMap<u8, Client>,
}

impl TcpServer {
    pub fn new(addr: SocketAddr) -> io::Result<Self> {
        let listener = TcpListener::bind(addr)?;
        listener.set_nonblocking(true)?;

        let listener_fd = listener.as_raw_fd();

        let epoll = Epoll::create()?;
        epoll.add_read_interest(listener_fd, KEY_NEW_CONNECTION)?;

        Ok(TcpServer {
            addr,
            listener,
            listener_fd,
            epoll,
            request_contexts: HashMap::new(),
        })
    }

    pub fn listen(&mut self) -> io::Result<()> {
        println!("Listening on {}", self.addr);

        loop {
            self.epoll.wait().expect("Failed to wait for epoll event");

            let events = &self.epoll.events.iter()
                .map(|event| (event.events, event.u64))
                .collect::<Vec<(u32, u64)>>();

            for (events, u64) in events {
                match *u64 as u16 {
                    KEY_NEW_CONNECTION => self.accept_connection()?,
                    key => self.handle_event(*events, key as u8)
                }
            }
        }
    }

    fn accept_connection(&mut self) -> io::Result<()> {
        match self.listener.accept() {
            Ok((stream, addr)) => {
                stream.set_nonblocking(true)?;

                let fd = stream.as_raw_fd();
                let client = Client::new(stream);
                println!("New client: {addr} ({})", client.id);

                self.epoll.add_read_interest(fd, client.id as u16)?;
                self.request_contexts.insert(client.id, client);
            }
            Err(e) => eprintln!("Couldn't accept: {e}")
        };

        self.epoll.modify_read_interest(self.listener_fd, KEY_NEW_CONNECTION)
    }

    fn handle_event(&mut self, events: u32, key: u8) {
        let mut to_delete = None;
        if let Some(client) = self.request_contexts.get_mut(&key) {
            match events {
                v if is_read_event(v) => {
                    let mut buf = [0u8; 1024];
                    let read_length = client.stream.read(&mut buf).expect("Failed to read stream");
                    let registration = ClientRegistration::deserialize(&buf[..read_length]);

                    println!("Test");
                }
                v if is_write_event(v) => {
                    println!("Write Event");
                    // context.write_cb(key, epoll_fd)?;
                    to_delete = Some(key);
                }
                v => println!("Unexpected event: {v}"),
            }
        }
        if let Some(key) = to_delete {
            self.request_contexts.remove(&key);
        }
    }
}
